from autolens.data.array.util import array_util
from autolens.data.array import scaled_array
from autolens.plotters import array_plotters

import os

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

lens_name = "SLACSJ0252+0039"
lens_name = "SLACSJ1250+0523"
# lens_name = 'SLACSJ1430+4105'

resi = array_util.numpy_array_2d_from_fits(
    file_path=path + "../data/slacs/slacs1250+0523/F814W_image.fits", hdu=0
)
resi = scaled_array.ScaledSquarePixelArray(array=resi, pixel_scale=0.03)
resi = resi.resized_scaled_array_from_array(new_shape=(201, 201))
# resi = np.flipud(resi)
array_plotters.plot_array(
    array=resi,
    norm="log",
    norm_min=0.0,
    norm_max=1.0,
    title="Observed CCD",
    output_path=path + "slacs_residuals_output/",
    output_filename=lens_name + "_Obs",
    output_format="png",
)

resi = array_util.numpy_array_2d_from_fits(
    file_path=path + "slacs_residuals/" + lens_name + "_Decomp_Mass_x1Sersic.fits",
    hdu=0,
)
resi = scaled_array.ScaledSquarePixelArray(array=resi, pixel_scale=0.03)
resi = resi.resized_scaled_array_from_array(new_shape=(201, 201))
# resi = np.flipud(resi)
array_plotters.plot_array(
    array=resi,
    norm="symmetric_log",
    norm_min=-0.1,
    norm_max=0.1,
    linthresh=0.0005,
    linscale=0.01,
    title="CCD Residuals (Sersic)",
    output_path=path + "slacs_residuals_output/",
    output_filename=lens_name + "_x1Sersic",
    output_format="png",
)

resi = array_util.numpy_array_2d_from_fits(
    file_path=path
    + "slacs_residuals/"
    + lens_name
    + "_Decomp_Mass_x2SersicOffset.fits",
    hdu=0,
)
resi = scaled_array.ScaledSquarePixelArray(array=resi, pixel_scale=0.03)
resi = resi.resized_scaled_array_from_array(new_shape=(201, 201))
# resi = np.flipud(resi)
array_plotters.plot_array(
    array=resi,
    norm="symmetric_log",
    norm_min=-0.1,
    norm_max=0.1,
    linthresh=0.0005,
    linscale=0.01,
    title="CCD Residuals (Sersic + Exp)",
    output_path=path + "slacs_residuals_output/",
    output_filename=lens_name + "_x2Sersic",
    output_format="png",
)

# Put in array plotters

# cb = plt.colorbar(fraction=cb_fraction, pad=cb_pad, ticks=[0.1, 0.05, 0.01, 0.001, 0.0, -0.001, -0.01, -0.05, -0.1])
# cb.ax.tick_params(labelsize=cb_ticksize)
# cb.ax.set_yticklabels(['< -0.1', '0.05', '0.01', '0.001', '0.0', '-0.001', '-0.01', '-0.05', '> 0.0'])



from test.integration.tests.full_pipeline import hyper_no_lens_light_bg
from test.integration.tests.runner import run_a_mock


class TestCase:
    def _test_hyper_with_lens_light(self):
        run_a_mock(hyper_no_lens_light_bg)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.inversion import pixelizations as pix, regularization as reg
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from test.integration.tests import runner

test_type = "lens_and_source_inversion"
test_name = "lens_mass_x1_source_x1_adaptive_brightness_hyper"
data_type = "no_lens_light__source_smooth"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class Phase1(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.source.light.sersic_index = af.UniformPrior(3.9, 4.1)

    phase1 = Phase1(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal),
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 50
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(hyper_galaxy=True)

    class InversionPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.last.constant.galaxies.lens

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.galaxies.source.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
            )

    phase2 = InversionPhase(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal),
            source=gm.GalaxyModel(
                redshift=1.0,
                pixelization=pix.VoronoiBrightnessImage,
                regularization=reg.AdaptiveBrightness,
            ),
        ),
        inversion_pixel_limit=50,
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.8

    phase2 = phase2.extend_with_multiple_hyper_phases(hyper_galaxy=True, inversion=True)

    class InversionPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.galaxies.lens = results.from_phase("phase_1").variable.galaxies.lens

            self.galaxies = results.last.hyper_combined.constant.galaxies

            self.galaxies.lens.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.lens.hyper_galaxy
            )

            self.galaxies.source.hyper_galaxy = (
                results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
            )

    phase3 = InversionPhase(
        phase_name="phase_3",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal),
            source=gm.GalaxyModel(
                redshift=1.0,
                pixelization=pix.VoronoiBrightnessImage,
                regularization=reg.AdaptiveBrightness,
            ),
        ),
        inversion_pixel_limit=50,
        optimizer_class=optimizer_class,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 40
    phase3.optimizer.sampling_efficiency = 0.8

    phase3 = phase3.extend_with_multiple_hyper_phases(hyper_galaxy=True, inversion=True)

    return pl.PipelineImaging(name, phase1, phase2, phase3)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from test.integration.tests.lens_and_source import lens_mass_x1_source_x2_hyper
from test.integration.tests.runner import run_a_mock


class TestCase:
    def _test_lens_mass_x1_source_x2_hyper(self):
        run_a_mock(lens_mass_x1_source_x2_hyper)



import numpy as np
from astropy import cosmology as cosmo
from functools import wraps

from autolens import exc
from autolens.data.array import grids
from autolens.lens import plane as pl
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.model.inversion import inversions as inv
from autolens.model.galaxy import galaxy as g

from autolens.data.array.grids import (
    reshape_returned_array,
    reshape_returned_array_blurring,
    reshape_returned_grid,
)


class AbstractTracerCosmology(object):
    def __init__(self, plane_redshifts, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        plane_redshifts : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """
        self.plane_redshifts = plane_redshifts
        self.cosmology = cosmology

    @property
    def total_planes(self):
        return len(self.plane_redshifts)

    def arcsec_per_kpc_proper_of_plane(self, i):
        return cosmology_util.arcsec_per_kpc_from_redshift_and_cosmology(
            redshift=self.plane_redshifts[i], cosmology=self.cosmology
        )

    def kpc_per_arcsec_proper_of_plane(self, i):
        return 1.0 / self.arcsec_per_kpc_proper_of_plane(i=i)

    def angular_diameter_distance_of_plane_to_earth_in_units(
        self, i, unit_length="arcsec"
    ):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.plane_redshifts[i],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def angular_diameter_distance_between_planes_in_units(
        self, i, j, unit_length="arcsec"
    ):
        return cosmology_util.angular_diameter_distance_between_redshifts_from_redshifts_and_cosmlology(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def angular_diameter_distance_to_source_plane_in_units(self, unit_length="arcsec"):
        return cosmology_util.angular_diameter_distance_to_earth_from_redshift_and_cosmology(
            redshift=self.plane_redshifts[-1],
            cosmology=self.cosmology,
            unit_length=unit_length,
        )

    def critical_surface_density_between_planes_in_units(
        self, i, j, unit_length="arcsec", unit_mass="solMass"
    ):
        return cosmology_util.critical_surface_density_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            cosmology=self.cosmology,
            unit_length=unit_length,
            unit_mass=unit_mass,
        )

    def scaling_factor_between_planes(self, i, j):
        return cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
            redshift_0=self.plane_redshifts[i],
            redshift_1=self.plane_redshifts[j],
            redshift_final=self.plane_redshifts[-1],
            cosmology=self.cosmology,
        )

    def angular_diameter_distance_from_image_to_source_plane_in_units(
        self, unit_length="arcsec"
    ):
        return self.angular_diameter_distance_between_planes_in_units(
            i=0, j=-1, unit_length=unit_length
        )

    def padded_tracer_from_psf_shape(self, psf_shape):
        raise NotImplementedError()


class AbstractTracer(AbstractTracerCosmology):
    def __init__(self, planes, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        planes : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        self.planes = planes
        super(AbstractTracer, self).__init__(
            plane_redshifts=[plane.redshift for plane in self.planes],
            cosmology=cosmology,
        )

    def plane_with_galaxy(self, galaxy):
        return [plane for plane in self.planes if galaxy in plane.galaxies][0]

    @property
    def grid_stack(self):
        return self.planes[0].grid_stack

    @property
    def image_plane(self):
        return self.planes[0]

    @property
    def source_plane(self):
        return self.planes[-1]

    @property
    def all_planes_have_redshifts(self):
        return None not in self.plane_redshifts

    @property
    def has_light_profile(self):
        return any(list(map(lambda plane: plane.has_light_profile, self.planes)))

    @property
    def has_mass_profile(self):
        return any(list(map(lambda plane: plane.has_mass_profile, self.planes)))

    @property
    def has_pixelization(self):
        return any(list(map(lambda plane: plane.has_pixelization, self.planes)))

    @property
    def has_regularization(self):
        return any(list(map(lambda plane: plane.has_regularization, self.planes)))

    @property
    def has_hyper_galaxy(self):
        return any(list(map(lambda plane: plane.has_hyper_galaxy, self.planes)))

    @property
    def galaxies(self):
        return list([galaxy for plane in self.planes for galaxy in plane.galaxies])

    @reshape_returned_array
    def profile_image_plane_image(self, return_in_2d=True, return_binned=True):
        return sum(
            self.profile_image_plane_image_of_planes(
                return_in_2d=False, return_binned=False
            )
        )

    def profile_image_plane_image_of_planes(
        self, return_in_2d=True, return_binned=True
    ):
        return [
            plane.profile_image_plane_image(
                return_in_2d=return_in_2d, return_binned=return_binned
            )
            for plane in self.planes
        ]

    def padded_profile_image_plane_image_2d_from_psf_shape(self, psf_shape):

        padded_tracer = self.padded_tracer_from_psf_shape(psf_shape=psf_shape)

        return padded_tracer.profile_image_plane_image(
            return_in_2d=True, return_binned=True
        )

    @reshape_returned_array_blurring
    def profile_image_plane_blurring_image(self, return_in_2d=True):
        return sum(
            self.profile_image_plane_blurring_image_of_planes(return_in_2d=False)
        )

    def profile_image_plane_blurring_image_of_planes(self, return_in_2d=True):
        return [
            plane.profile_image_plane_blurring_image(return_in_2d=return_in_2d)
            for plane in self.planes
        ]

    @property
    def mappers_of_planes(self):
        return list(filter(None, [plane.mapper for plane in self.planes]))

    @property
    def regularizations_of_planes(self):
        return list(filter(None, [plane.regularization for plane in self.planes]))

    @reshape_returned_array
    def convergence(self, return_in_2d=True, return_binned=True):
        return sum(
            [
                plane.convergence(return_in_2d=False, return_binned=False)
                for plane in self.planes
            ]
        )

    @reshape_returned_array
    def potential(self, return_in_2d=True, return_binned=True):
        return sum(
            [
                plane.potential(return_in_2d=False, return_binned=False)
                for plane in self.planes
            ]
        )

    @reshape_returned_array
    def deflections_y(self, return_in_2d=True, return_binned=True):
        return sum(
            [
                plane.deflections_y(return_in_2d=False, return_binned=False)
                for plane in self.planes
            ]
        )

    @reshape_returned_array
    def deflections_x(self, return_in_2d=True, return_binned=True):
        return sum(
            [
                plane.deflections_x(return_in_2d=False, return_binned=False)
                for plane in self.planes
            ]
        )

    @reshape_returned_grid
    def deflections(self, return_in_2d=True, return_binned=True):
        return sum(
            [
                plane.deflections(return_in_2d=False, return_binned=False)
                for plane in self.planes
            ]
        )

    def einstein_radius_of_plane_in_units(self, i, unit_length="arcsec"):
        return self.planes[i].einstein_radius_in_units(unit_length=unit_length)

    def einstein_mass_between_planes_in_units(self, i, j, unit_mass="solMass"):
        return self.planes[i].einstein_mass_in_units(
            unit_mass=unit_mass, redshift_source=self.plane_redshifts[j]
        )

    def grid_at_redshift_from_image_plane_grid_and_redshift(
        self, image_plane_grid, redshift
    ):
        """For an input grid of (y,x) arc-second image-plane coordinates, ray-trace the coordinates to any redshift in \
        the strong lens configuration.

        This is performed using multi-plane ray-tracing and the existing redshifts and planes of the tracer. However, \
        any redshift can be input even if a plane does not exist there, including redshifts before the first plane \
        of the lensing system.

        Parameters
        ----------
        image_plane_grid : ndsrray or grids.Grid
            The image-plane grid which is traced to the redshift.
        redshift : float
            The redshift the image-plane grid is traced to.
        """

        # TODO : We need to come up with a better abstraction for multi-plane lensing 0_0

        image_plane_grid_stack = grids.GridStack(
            regular=image_plane_grid,
            sub=np.array([[0.0, 0.0]]),
            blurring=np.array([[0.0, 0.0]]),
        )

        tracer = Tracer.from_galaxies_and_image_plane_grid_stack(
            galaxies=self.galaxies,
            image_plane_grid_stack=image_plane_grid_stack,
            border=None,
            cosmology=self.cosmology,
        )

        for plane_index in range(0, len(self.plane_redshifts)):

            new_grid_stack = image_plane_grid_stack

            if redshift <= tracer.plane_redshifts[plane_index]:

                # If redshift is between two planes, we need to map over all previous planes coordinates / deflections.

                if plane_index > 0:
                    for previous_plane_index in range(plane_index):
                        scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                            redshift_0=tracer.plane_redshifts[previous_plane_index],
                            redshift_1=redshift,
                            redshift_final=tracer.plane_redshifts[-1],
                            cosmology=tracer.cosmology,
                        )

                        scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                            plane=tracer.planes[previous_plane_index],
                            scaling_factor=scaling_factor,
                        )

                        new_grid_stack = lens_util.grid_stack_from_deflections_stack(
                            grid_stack=new_grid_stack,
                            deflections_stack=scaled_deflections_stack,
                        )

                # If redshift is before the first plane, no change to image pllane coordinates.

                elif plane_index == 0:

                    return new_grid_stack.regular

                return new_grid_stack.regular


class AbstractTracerData(AbstractTracer):
    def __init__(self, planes, cosmology):
        """Abstract Ray tracer for lens systems with any number of planes.

        From the galaxies of the tracer's planes, their grid-stack(s) and the cosmology physically derived quantities \
        (e.g. surface density, angular diameter distances, critical surface densities) can be computed.

        Parameters
        ----------
        planes : [pl.Plane]
            The list of the tracer's planes in ascending redshift order.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """
        super(AbstractTracerData, self).__init__(planes=planes, cosmology=cosmology)

    def blurred_profile_image_plane_image_2d_of_planes_from_convolver_image(
        self, convolver_image
    ):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution) and then map them back to the 2D array of the original mask.

        The blurred image of every plane is returned in 2D.

        Parameters
        ----------
        convolver_image : hyper_galaxy.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        blurred_profile_image_plane_images_of_planes_1d = self.blurred_profile_image_plane_image_1d_of_planes_from_convolver_image(
            convolver_image=convolver_image
        )

        return list(
            map(
                lambda blurred_profile_image_plane_image_1d: self.image_plane.grid_stack.scaled_array_2d_from_array_1d(
                    array_1d=blurred_profile_image_plane_image_1d
                ),
                blurred_profile_image_plane_images_of_planes_1d,
            )
        )

    def blurred_profile_image_plane_image_1d_from_convolver_image(
        self, convolver_image
    ):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution).

        These are summed to give the tracer's overall blurred image-plane image in 1D.

        Parameters
        ----------
        convolver_image : hyper_galaxy.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """

        image_array = self.profile_image_plane_image(
            return_in_2d=False, return_binned=True
        )
        blurring_array = self.profile_image_plane_blurring_image(return_in_2d=False)

        return convolver_image.convolve_image(
            image_array=image_array, blurring_array=blurring_array
        )

    def blurred_profile_image_plane_image_1d_of_planes_from_convolver_image(
        self, convolver_image
    ):
        """Extract the 1D image-plane image and 1D blurring image-plane image of every plane and blur each with the \
        PSF using a convolver (see ccd.convolution).

        The blurred image of every plane is returned in 1D.

        Parameters
        ----------
        convolver_image : hyper_galaxy.ccd.convolution.ConvolverImage
            Class which performs the PSF convolution of a masked image in 1D.
        """
        return [
            plane.blurred_profile_image_plane_image_1d_from_convolver_image(
                convolver_image=convolver_image
            )
            for plane in self.planes
        ]

    def unmasked_blurred_profile_image_plane_image_from_psf(self, psf):

        padded_tracer = self.padded_tracer_from_psf_shape(psf_shape=psf.shape)

        padded_image_1d = padded_tracer.profile_image_plane_image(
            return_in_2d=False, return_binned=True
        )

        return padded_tracer.grid_stack.unmasked_blurred_array_2d_from_padded_array_1d_psf_and_image_shape(
            padded_array_1d=padded_image_1d,
            psf=psf,
            image_shape=self.grid_stack.regular.mask.shape,
        )

    def unmasked_blurred_profile_image_plane_image_of_planes_from_psf(self, psf):

        unmasked_blurred_profile_image_plane_image_of_planes = []

        padded_tracer = self.padded_tracer_from_psf_shape(psf_shape=psf.shape)

        for padded_plane in padded_tracer.planes:

            padded_image_1d = padded_plane.profile_image_plane_image(
                return_in_2d=False, return_binned=True
            )

            unmasked_blurred_array_2d = padded_tracer.grid_stack.unmasked_blurred_array_2d_from_padded_array_1d_psf_and_image_shape(
                padded_array_1d=padded_image_1d,
                psf=psf,
                image_shape=self.grid_stack.regular.mask.shape,
            )

            unmasked_blurred_profile_image_plane_image_of_planes.append(
                unmasked_blurred_array_2d
            )

        return unmasked_blurred_profile_image_plane_image_of_planes

    def unmasked_blurred_profile_image_plane_image_of_plane_and_galaxies_from_psf(
        self, psf
    ):

        unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies = []

        padded_tracer = self.padded_tracer_from_psf_shape(psf_shape=psf.shape)

        for padded_plane in padded_tracer.planes:

            padded_image_1d_of_galaxies = padded_plane.profile_image_plane_image_of_galaxies(
                return_in_2d=False, return_binned=True
            )

            unmasked_blurred_array_2d_of_galaxies = list(
                map(
                    lambda padded_image_1d_of_galaxy: padded_tracer.grid_stack.unmasked_blurred_array_2d_from_padded_array_1d_psf_and_image_shape(
                        padded_array_1d=padded_image_1d_of_galaxy,
                        psf=psf,
                        image_shape=self.grid_stack.regular.mask.shape,
                    ),
                    padded_image_1d_of_galaxies,
                )
            )

            unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies.append(
                unmasked_blurred_array_2d_of_galaxies
            )

        return unmasked_blurred_profile_image_plane_image_of_planes_and_galaxies

    def inversion_from_image_1d_noise_map_1d_and_convolver_mapping_matrix(
        self, image_1d, noise_map_1d, convolver_mapping_matrix
    ):

        if len(self.mappers_of_planes) > 1:
            raise exc.RayTracingException(
                "PyAutoLens does not currently support more than one mapper, reglarization and"
                "therefore inversion per tracer."
            )

        return inv.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=image_1d,
            noise_map_1d=noise_map_1d,
            convolver=convolver_mapping_matrix,
            mapper=self.mappers_of_planes[0],
            regularization=self.regularizations_of_planes[0],
        )

    def hyper_noise_map_1d_from_noise_map_1d(self, noise_map_1d):
        hyper_noise_maps_1d = self.hyper_noise_maps_1d_of_planes_from_noise_map_1d(
            noise_map_1d=noise_map_1d
        )
        hyper_noise_maps_1d = [
            hyper_noise_map
            for hyper_noise_map in hyper_noise_maps_1d
            if hyper_noise_map is not None
        ]
        return sum(hyper_noise_maps_1d)

    def hyper_noise_maps_1d_of_planes_from_noise_map_1d(self, noise_map_1d):
        return [
            plane.hyper_noise_map_1d_from_noise_map_1d(noise_map_1d=noise_map_1d)
            for plane in self.planes
        ]

    @property
    def galaxy_image_dict_blank_images(self) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        for plane in self.planes:
            for galaxy in plane.galaxies:

                galaxy_image_dict[galaxy] = plane.profile_image_plane_image_of_galaxy(
                    galaxy=galaxy, return_in_2d=False, return_binned=True
                )

        return galaxy_image_dict

    def galaxy_image_dict_from_convolver_image(
        self, convolver_image
    ) -> {g.Galaxy: np.ndarray}:
        """
        A dictionary associating galaxies with their corresponding model images
        """

        galaxy_image_dict = dict()

        for plane in self.planes:
            for galaxy in plane.galaxies:

                profile_image_plane_image_1d = plane.profile_image_plane_image_of_galaxy(
                    galaxy=galaxy, return_in_2d=False, return_binned=True
                )

                profile_image_plane_blurring_image_1d = plane.profile_image_plane_blurring_image_of_galaxy(
                    galaxy=galaxy, return_in_2d=False
                )

                blurred_profile_image_pane_image_1d = convolver_image.convolve_image(
                    image_array=profile_image_plane_image_1d,
                    blurring_array=profile_image_plane_blurring_image_1d,
                )

                galaxy_image_dict[galaxy] = blurred_profile_image_pane_image_1d

        return galaxy_image_dict


class Tracer(AbstractTracerData):
    def __init__(self, planes, cosmology=cosmo.Planck15):
        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the redshits of the galaxies; there is a unique plane redshift \
        for every unique galaxy redshift (galaxies with identical redshifts are put in the same plane).

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        super(Tracer, self).__init__(planes=planes, cosmology=cosmology)

    @classmethod
    def from_galaxies_and_image_plane_grid_stack(
        cls, galaxies, image_plane_grid_stack, border=None, cosmology=cosmo.Planck15
    ):

        plane_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(
            galaxies=galaxies
        )

        if len(plane_redshifts) == 1:
            return cls.x1_plane_tracer_from_lens_galaxies_and_image_plane_grid_stack(
                lens_galaxies=galaxies,
                image_plane_grid_stack=image_plane_grid_stack,
                border=border,
                cosmology=cosmology,
            )

        galaxies_in_planes = lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(
            galaxies=galaxies, plane_redshifts=plane_redshifts
        )

        if len(plane_redshifts) == 2:
            return cls.x2_plane_tracer_from_lens_and_source_galaxies_and_image_plane_grid_stack(
                lens_galaxies=galaxies_in_planes[0],
                source_galaxies=galaxies_in_planes[1],
                image_plane_grid_stack=image_plane_grid_stack,
                border=border,
                cosmology=cosmology,
            )

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            compute_deflections = lens_util.compute_deflections_at_next_plane(
                plane_index=plane_index, total_planes=len(plane_redshifts)
            )

            new_grid_stack = image_plane_grid_stack

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index],
                        redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1],
                        cosmology=cosmology,
                    )

                    scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                        plane=planes[previous_plane_index],
                        scaling_factor=scaling_factor,
                    )

                    new_grid_stack = lens_util.grid_stack_from_deflections_stack(
                        grid_stack=new_grid_stack,
                        deflections_stack=scaled_deflections_stack,
                    )

            planes.append(
                pl.Plane(
                    galaxies=galaxies_in_planes[plane_index],
                    grid_stack=new_grid_stack,
                    border=border,
                    compute_deflections=compute_deflections,
                    cosmology=cosmology,
                )
            )

        return Tracer(planes=planes, cosmology=cosmology)

    @classmethod
    def x1_plane_tracer_from_lens_galaxies_and_image_plane_grid_stack(
        cls,
        lens_galaxies,
        image_plane_grid_stack,
        border=None,
        cosmology=cosmo.Planck15,
    ):
        image_plane = pl.Plane(
            galaxies=lens_galaxies,
            grid_stack=image_plane_grid_stack,
            border=border,
            compute_deflections=False,
            cosmology=cosmology,
        )

        return Tracer(planes=[image_plane], cosmology=cosmology)

    @classmethod
    def x2_plane_tracer_from_lens_and_source_galaxies_and_image_plane_grid_stack(
        cls,
        lens_galaxies,
        source_galaxies,
        image_plane_grid_stack,
        border=None,
        cosmology=cosmo.Planck15,
    ):
        image_plane = pl.Plane(
            galaxies=lens_galaxies,
            grid_stack=image_plane_grid_stack,
            border=border,
            compute_deflections=True,
            cosmology=cosmology,
        )

        source_plane_grid_stack = image_plane.trace_grid_stack_to_next_plane()

        source_plane = pl.Plane(
            galaxies=source_galaxies,
            grid_stack=source_plane_grid_stack,
            border=border,
            compute_deflections=False,
            cosmology=cosmology,
        )
        return Tracer(planes=[image_plane, source_plane], cosmology=cosmology)

    @classmethod
    def sliced_tracer_from_lens_line_of_sight_and_source_galaxies(
        cls,
        lens_galaxies,
        line_of_sight_galaxies,
        source_galaxies,
        planes_between_lenses,
        image_plane_grid_stack,
        border=None,
        cosmology=cosmo.Planck15,
    ):

        """Ray-tracer for a lens system with any number of planes.

        The redshift of these planes are specified by the input parameters *lens_redshifts* and \
         *slices_between_main_planes*. Every galaxy is placed in its closest plane in redshift-space.

        To perform multi-plane ray-tracing, a cosmology must be supplied so that deflection-angles can be rescaled \
        according to the lens-geometry of the multi-plane system. All galaxies input to the tracer must therefore \
        have redshifts.

        This tracer has only one grid-stack (see grid_stack.GridStack) which is used for ray-tracing.

        Parameters
        ----------
        lens_galaxies : [Galaxy]
            The list of galaxies in the ray-tracing calculation.
        image_plane_grid_stack : grid_stacks.GridStack
            The image-plane grid stack which is traced. (includes the regular-grid, sub-grid, blurring-grid, etc.).
        planes_between_lenses : [int]
            The number of slices between each main plane. The first entry in this list determines the number of slices \
            between Earth (redshift 0.0) and main plane 0, the next between main planes 0 and 1, etc.
        border : masks.RegularGridBorder
            The border of the regular-grid, which is used to relocate demagnified traced pixels to the \
            source-plane borders.
        cosmology : astropy.cosmology
            The cosmology of the ray-tracing calculation.
        """

        lens_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(
            galaxies=lens_galaxies
        )

        plane_redshifts = lens_util.ordered_plane_redshifts_from_lens_and_source_plane_redshifts_and_slice_sizes(
            lens_redshifts=lens_redshifts,
            planes_between_lenses=planes_between_lenses,
            source_plane_redshift=source_galaxies[0].redshift,
        )

        galaxies_in_planes = lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(
            galaxies=lens_galaxies + line_of_sight_galaxies,
            plane_redshifts=plane_redshifts,
        )

        plane_redshifts.append(source_galaxies[0].redshift)
        galaxies_in_planes.append(source_galaxies)

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            compute_deflections = lens_util.compute_deflections_at_next_plane(
                plane_index=plane_index, total_planes=len(plane_redshifts)
            )

            new_grid_stack = image_plane_grid_stack

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index],
                        redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1],
                        cosmology=cosmology,
                    )

                    scaled_deflections_stack = lens_util.scaled_deflections_stack_from_plane_and_scaling_factor(
                        plane=planes[previous_plane_index],
                        scaling_factor=scaling_factor,
                    )

                    new_grid_stack = lens_util.grid_stack_from_deflections_stack(
                        grid_stack=new_grid_stack,
                        deflections_stack=scaled_deflections_stack,
                    )

            planes.append(
                pl.Plane(
                    redshift=plane_redshifts[plane_index],
                    galaxies=galaxies_in_planes[plane_index],
                    grid_stack=new_grid_stack,
                    border=border,
                    compute_deflections=compute_deflections,
                    cosmology=cosmology,
                )
            )

        return Tracer(planes=planes, cosmology=cosmology)

    @classmethod
    def from_galaxies_and_image_plane_positions(
        cls, galaxies, image_plane_positions, cosmology=cosmo.Planck15
    ):

        plane_redshifts = lens_util.ordered_plane_redshifts_from_galaxies(
            galaxies=galaxies
        )

        galaxies_in_planes = lens_util.galaxies_in_redshift_ordered_planes_from_galaxies(
            galaxies=galaxies, plane_redshifts=plane_redshifts
        )

        planes = []

        for plane_index in range(0, len(plane_redshifts)):

            compute_deflections = lens_util.compute_deflections_at_next_plane(
                plane_index=plane_index, total_planes=len(plane_redshifts)
            )

            new_positions = image_plane_positions

            if plane_index > 0:
                for previous_plane_index in range(plane_index):
                    scaling_factor = cosmology_util.scaling_factor_between_redshifts_from_redshifts_and_cosmology(
                        redshift_0=plane_redshifts[previous_plane_index],
                        redshift_1=plane_redshifts[plane_index],
                        redshift_final=plane_redshifts[-1],
                        cosmology=cosmology,
                    )

                    scaled_deflections = list(
                        map(
                            lambda deflections: np.multiply(
                                scaling_factor, deflections
                            ),
                            planes[previous_plane_index].deflections,
                        )
                    )

                    new_positions = list(
                        map(
                            lambda positions, deflections: np.subtract(
                                positions, deflections
                            ),
                            new_positions,
                            scaled_deflections,
                        )
                    )

            planes.append(
                pl.PlanePositions(
                    redshift=plane_redshifts[plane_index],
                    galaxies=galaxies_in_planes[plane_index],
                    positions=new_positions,
                    compute_deflections=compute_deflections,
                )
            )

        return Tracer(planes=planes, cosmology=cosmology)

    def padded_tracer_from_psf_shape(self, psf_shape):

        padded_grid_stack = self.grid_stack.padded_grid_stack_from_psf_shape(
            psf_shape=psf_shape
        )

        return Tracer.from_galaxies_and_image_plane_grid_stack(
            galaxies=self.galaxies,
            image_plane_grid_stack=padded_grid_stack,
            border=self.image_plane.border,
            cosmology=self.cosmology,
        )



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration.tests import runner

test_type = "model_mapper"
test_name = "constants_x2_profile"
data_type = "lens_only_dev_vaucouleurs"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class MMPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens.light_0.axis_ratio = 0.2
            self.galaxies.lens.light_0.phi = 90.0
            self.galaxies.lens.light_0.centre_0 = 1.0
            self.galaxies.lens.light_0.centre_1 = 2.0
            self.galaxies.lens.light_1.axis_ratio = 0.2
            self.galaxies.lens.light_1.phi = 90.0
            self.galaxies.lens.light_1.centre_0 = 1.0
            self.galaxies.lens.light_1.centre_1 = 2.0

    phase1 = MMPhase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, light_0=lp.EllipticalSersic, light_1=lp.EllipticalSersic
            )
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(name, phase1)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import numpy as np
import pytest

from autolens.data.array import grids, mask as msk
from autolens.model.inversion import mappers as m
from autolens.model.inversion import pixelizations, regularization
from test.unit.mock.data.mock_grids import MockPixSubGrid, MockPixGridStack


class TestRectangular:
    def test__5_simple_grid__no_sub_grid(self):

        # Source-plane comprises 5 grid, so 5 masked_image pixels traced to the pix-plane.
        regular_grid = np.array(
            [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
        )
        sub_grid = np.array(
            [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
        )

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(
                sub_grid=sub_grid, sub_to_regular=sub_to_regular, sub_grid_size=1
            ),
        )

        # There is no sub-grid, so our sub_grid are just the masked_image grid (note the NumPy weighted_data structure
        # ensures this has no sub-gridding)

        pix = pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None, hyper_image=np.ones((2, 2))
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)
        assert (mapper.hyper_image == np.ones((2, 2))).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            pixel_neighbors=mapper.geometry.pixel_neighbors,
            pixel_neighbors_size=mapper.geometry.pixel_neighbors_size,
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__15_grid__no_sub_grid(self):

        # Source-plane comprises 15 grid, so 15 masked_image pixels traced to the pix-plane.

        regular_grid = np.array(
            [
                [0.9, -0.9],
                [1.0, -1.0],
                [1.1, -1.1],
                [0.9, 0.9],
                [1.0, 1.0],
                [1.1, 1.1],
                [0.01, 0.01],
                [0.0, 0.0],
                [0.01, 0.01],
                [-0.9, -0.9],
                [-1.0, -1.0],
                [-1.1, -1.1],
                [-0.9, 0.9],
                [-1.0, 1.0],
                [-1.1, 1.1],
            ]
        )

        # There is no sub-grid, so our sub_grid are just the masked_image grid (note the NumPy weighted_data structure
        # ensures this has no sub-gridding)
        sub_grid = np.array(
            [
                [0.9, -0.9],
                [1.0, -1.0],
                [1.1, -1.1],
                [0.9, 0.9],
                [1.0, 1.0],
                [1.1, 1.1],
                [-0.01, 0.01],
                [0.0, 0.0],
                [0.01, 0.01],
                [-0.9, -0.9],
                [-1.0, -1.0],
                [-1.1, -1.1],
                [-0.9, 0.9],
                [-1.0, 1.0],
                [-1.1, 1.1],
            ]
        )

        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
        )

        pix = pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.2, 2.2), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__5_simple_grid__include_sub_grid(self):
        # Source-plane comprises 5 grid, so 5 masked_image pixels traced to the pix-plane.
        regular_grid = np.array(
            [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
        )
        # Assume a 2x2 sub-grid, so each of our 5 masked_image-pixels are split into 4.
        # The grid below is unphysical in that the (0.0, 0.0) terms on the end of each sub-grid probably couldn't
        # happen for a real lensing calculation. This is to make a mapping_matrix matrix which explicitly tests the
        # sub-grid.
        sub_grid = np.array(
            [
                [1.0, -1.0],
                [1.0, -1.0],
                [1.0, -1.0],
                [0.0, 0.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [-1.0, -1.0],
                [-1.0, -1.0],
                [-1.0, -1.0],
                [0.0, 0.0],
                [-1.0, 1.0],
                [-1.0, 1.0],
                [-1.0, 1.0],
                [0.0, 0.0],
            ]
        )

        sub_to_regular = np.array(
            [0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2]
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=2),
        )

        pix = pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.75, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.75],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__grid__requires_border_relocation(self):

        regular_grid = np.array(
            [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
        )
        sub_grid = np.array(
            [[2.0, 2.0], [2.0, 2.0], [2.0, 2.0], [2.0, 2.0], [-2.0, -2.0]]
        )

        border = grids.GridBorder(arr=np.array([0, 1, 3, 4]))

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
        )

        pix = pixelizations.Rectangular(shape=(3, 3))

        mapper = pix.mapper_from_grid_stack_and_border(grid_stack, border)

        assert mapper.is_image_plane_pixelization == False
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                ]
            )
        ).all()
        assert mapper.shape == (3, 3)

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()


class TestSparseToRegularGrid:
    def test__3x3_simple_grid__create_using_regular_grid(self):

        regular_grid = np.array(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-1.0, -1.0],
                [-1.0, 0.0],
                [-1.0, 1.0],
            ]
        )

        mask = msk.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        sub_grid = np.array(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, -1.0],
                [0.0, 0.0],
                [0.0, 1.0],
                [-1.0, -1.0],
                [-1.0, 0.0],
                [-1.0, 1.0],
            ]
        )
        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)
        sub_grid = MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1)

        pix = pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
            unmasked_sparse_shape=pix.shape, regular_grid=regular_grid
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=sub_grid,
            pix=sparse_to_regular_grid.sparse,
            regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse,
        )

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None, hyper_image=np.ones((2, 2))
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_regular_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)
        assert (mapper.hyper_image == np.ones((2, 2))).all()

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [2.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, -1.0, 2.00000001, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
                    [-1.0, 0.0, 0.0, 3.00000001, -1.0, 0.0, -1.0, 0.0, 0.0],
                    [0.0, -1.0, 0.0, -1.0, 4.00000001, -1.0, 0.0, -1.0, 0.0],
                    [0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, 0.0, 0.0, -1.0],
                    [0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 2.00000001, -1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, -1.0, 2.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask__create_using_regular_grid(self):

        regular_grid = np.array(
            [[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0], [-1.0, 0.0]]
        )

        mask = msk.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        sub_grid = np.array(
            [[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0], [-1.0, 0.0]]
        )
        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)
        sub_grid = MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1)

        pix = pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
            unmasked_sparse_shape=pix.shape, regular_grid=regular_grid
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=sub_grid,
            pix=sparse_to_regular_grid.sparse,
            regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse,
        )

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_regular_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.0), 1.0e-4)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask_and_sub_grid__create_using_regular_grid(
        self
    ):

        regular_grid = np.array(
            [[1.0, 0.0], [0.0, -1.0], [0.0, 0.0], [0.0, 1.0], [-1.0, 0.0]]
        )

        mask = msk.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        sub_grid = np.array(
            [
                [1.01, 0.0],
                [1.01, 0.0],
                [1.01, 0.0],
                [0.01, 0.0],
                [0.0, -1.0],
                [0.0, -1.0],
                [0.0, -1.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.01, 0.0],
                [0.0, 1.01],
                [0.0, 1.01],
                [0.0, 1.01],
                [0.01, 0.0],
                [-1.01, 0.0],
                [-1.01, 0.0],
                [-1.01, 0.0],
                [0.01, 0.0],
            ]
        )

        sub_to_regular = np.array(
            [0, 0, 0, 2, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 2, 4, 4, 4, 2]
        )

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)
        sub_grid = MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=2)

        pix = pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
            unmasked_sparse_shape=pix.shape, regular_grid=regular_grid
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=sub_grid,
            pix=sparse_to_regular_grid.sparse,
            regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse,
        )

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.02, 2.01), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_regular_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((0.0, 0.005), 1.0e-4)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.75, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.75, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.75, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.75],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask_with_offset_centre__create_using_regular_grid(
        self
    ):

        regular_grid = np.array(
            [[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]]
        )

        mask = msk.Mask(
            array=np.array(
                [
                    [True, True, True, False, True],
                    [True, True, False, False, False],
                    [True, True, True, False, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
        )

        sub_grid = np.array(
            [[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]]
        )
        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)
        sub_grid = MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1)

        pix = pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
            unmasked_sparse_shape=pix.shape, regular_grid=regular_grid
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=sub_grid,
            pix=sparse_to_regular_grid.sparse,
            regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse,
        )

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_regular_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((1.0, 1.0), 1.0e-4)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()


class TestVoronoiMagnification:
    def test__5_simple_grid__no_sub_grid(self):

        regular_grid = np.array(
            [[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]]
        )
        sub_grid = np.array(
            [[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]]
        )

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_to_sparse = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
            pix=regular_grid,
            regular_to_pixelization=regular_to_sparse,
        )

        pix = pixelizations.VoronoiMagnification(shape=(5, 1))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None, hyper_image=np.ones((2, 2))
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == regular_grid).all()
        assert mapper.geometry.origin == (0.0, 0.0)
        assert (mapper.hyper_image == np.ones((2, 2))).all()

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__3x3_simple_grid__include_mask_with_offset_centre__create_using_regular_grid(
        self
    ):

        regular_grid = np.array(
            [[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]]
        )

        mask = msk.Mask(
            array=np.array(
                [
                    [True, True, True, False, True],
                    [True, True, False, False, False],
                    [True, True, True, False, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
        )

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)

        sub_grid = np.array(
            [[2.0, 1.0], [1.0, 0.0], [1.0, 1.0], [1.0, 2.0], [0.0, 1.0]]
        )

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_to_sparse = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
            pix=regular_grid,
            regular_to_pixelization=regular_to_sparse,
        )

        pix = pixelizations.VoronoiMagnification(shape=(5, 1))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == regular_grid).all()
        assert mapper.geometry.origin == (1.0, 1.0)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_grid = grids.Grid(arr=regular_grid, mask=mask)
        sub_grid = MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1)

        pix = pixelizations.VoronoiMagnification(shape=(3, 3))
        sparse_to_regular_grid = grids.SparseToRegularGrid.from_unmasked_2d_grid_shape_and_regular_grid(
            unmasked_sparse_shape=pix.shape, regular_grid=regular_grid
        )

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=sub_grid,
            pix=sparse_to_regular_grid.sparse,
            regular_to_pixelization=sparse_to_regular_grid.regular_to_sparse,
        )

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == sparse_to_regular_grid.sparse).all()
        assert mapper.geometry.origin == pytest.approx((1.0, 1.0), 1.0e-4)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__15_grid__no_sub_grid(self):

        regular_grid = np.array(
            [
                [0.9, 0.9],
                [1.0, 1.0],
                [1.1, 1.1],
                [-0.9, 0.9],
                [-1.0, 1.0],
                [-1.1, 1.1],
                [-0.01, 0.01],
                [0.0, 0.0],
                [0.01, 0.01],
                [0.9, -0.9],
                [1.0, -1.0],
                [1.1, -1.1],
                [-0.9, -0.9],
                [-1.0, -1.0],
                [-1.1, -1.1],
            ]
        )

        sub_grid = np.array(
            [
                [0.9, 0.9],
                [1.0, 1.0],
                [1.1, 1.1],
                [-0.9, 0.9],
                [-1.0, 1.0],
                [-1.1, 1.1],
                [-0.01, 0.01],
                [0.0, 0.0],
                [0.01, 0.01],
                [0.9, -0.9],
                [1.0, -1.0],
                [1.1, -1.1],
                [-0.9, -0.9],
                [-1.0, -1.0],
                [-1.1, -1.1],
            ]
        )

        pixel_centers = np.array(
            [[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]]
        )

        sub_to_regular = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        regular_to_sparse = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
            pix=pixel_centers,
            regular_to_pixelization=regular_to_sparse,
        )

        pix = pixelizations.VoronoiMagnification(shape=(5, 1))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.2, 2.2), 1.0e-4)
        assert (mapper.geometry.pixel_centres == pixel_centers).all()
        assert mapper.geometry.origin == (0.0, 0.0)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__5_simple_grid__include_sub_grid__sets_up_correct_mapper(self):

        regular_grid = np.array(
            [[1.0, 1.0], [-1.0, 1.0], [0.0, 0.0], [1.0, -1.0], [-1.0, -1.0]]
        )

        sub_grid = np.array(
            [
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [0.0, 0.0],
                [-1.0, 1.0],
                [-1.0, 1.0],
                [-1.0, 1.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [1.0, -1.0],
                [1.0, -1.0],
                [1.0, -1.0],
                [0.0, 0.0],
                [-1.0, -1.0],
                [-1.0, -1.0],
                [-1.0, -1.0],
                [0.0, 0.0],
            ]
        )

        sub_to_regular = np.array(
            [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4]
        )

        pixel_centers = regular_grid

        regular_to_sparse = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=2),
            pix=pixel_centers,
            regular_to_pixelization=regular_to_sparse,
        )

        pix = pixelizations.VoronoiMagnification(shape=(5, 1))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=None
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert (mapper.geometry.pixel_centres == pixel_centers).all()
        assert mapper.geometry.origin == (0.0, 0.0)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.75, 0.0, 0.25, 0.0, 0.0],
                    [0.0, 0.75, 0.25, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.25, 0.75, 0.0],
                    [0.0, 0.0, 0.25, 0.0, 0.75],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()

    def test__same_as_above_but_grid_requires_border_relocation(self):

        regular_grid = np.array(
            [[1.0, -1.0], [1.0, 1.0], [0.0, 0.0], [-1.0, -1.0], [-1.0, 1.0]]
        )
        sub_grid = np.array(
            [[2.0, 2.0], [2.0, 2.0], [2.0, 2.0], [2.0, 2.0], [-2.0, -2.0]]
        )
        # These will all be relocated to the regular grid edge.
        pix_grid = np.array(
            [[1.1, -1.1], [1.1, 1.1], [0.0, 0.0], [-1.1, -1.1], [-1.1, 1.1]]
        )

        border = grids.GridBorder(arr=np.array([0, 1, 3, 4]))

        sub_to_regular = np.array([0, 1, 2, 3, 4])

        regular_to_sparse = np.array([0, 1, 2, 3, 4])

        grid_stack = MockPixGridStack(
            regular=regular_grid,
            sub=MockPixSubGrid(sub_grid, sub_to_regular, sub_grid_size=1),
            pix=pix_grid,
            regular_to_pixelization=regular_to_sparse,
        )

        pix = pixelizations.VoronoiMagnification(shape=(5, 1))

        mapper = pix.mapper_from_grid_stack_and_border(
            grid_stack=grid_stack, border=border
        )

        assert mapper.is_image_plane_pixelization == True
        assert mapper.geometry.shape_arcsec == pytest.approx((2.0, 2.0), 1.0e-4)
        assert mapper.geometry.pixel_centres == pytest.approx(regular_grid, 1e-4)
        assert mapper.geometry.origin == (0.0, 0.0)

        assert isinstance(mapper, m.VoronoiMapper)

        assert (
            mapper.mapping_matrix
            == np.array(
                [
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 0.0],
                ]
            )
        ).all()

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            mapper.geometry.pixel_neighbors, mapper.geometry.pixel_neighbors_size
        )

        assert (
            regularization_matrix
            == np.array(
                [
                    [3.00000001, -1.0, -1.0, -1.0, 0.0],
                    [-1.0, 3.00000001, -1.0, 0.0, -1.0],
                    [-1.0, -1.0, 4.00000001, -1.0, -1.0],
                    [-1.0, 0.0, -1.0, 3.00000001, -1.0],
                    [0.0, -1.0, -1.0, -1.0, 3.00000001],
                ]
            )
        ).all()



from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

from astropy.cosmology import FlatLambdaCDM


pixel_scale_img = 0.05
npixel_img = int(10 / pixel_scale_img)

H0 = 100.0
OM0 = 0.3
cosmo = FlatLambdaCDM(H0=H0, Om0=OM0)

psf = abstract_data.PSF.from_gaussian(
    shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale_img
)

image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
    shape=(200, 200), pixel_scale=pixel_scale_img, psf_shape=psf.shape, sub_grid_size=1
)

lens_galaxy = g.Galaxy(
    mass=mp.SphericalIsothermal(centre=(3.0, 3.0), einstein_radius=1.6), redshift=0.5
)

source_galaxy = g.Galaxy(
    light=lp.EllipticalSersic(
        centre=(0.1, 0.0),
        axis_ratio=0.8,
        phi=0.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
    redshift=1.0,
)

# "zero kappa" Mass Sheet
mass_sheet = g.Galaxy(sheet=mp.MassSheet(centre=(0.0, 0.0), kappa=0.0), redshift=0.2)

tracer_ref = ray_tracing.Tracer(
    galaxies=[lens_galaxy, source_galaxy],
    image_plane_grid_stack=image_plane_grid_stack,
    cosmology=cosmo,
)

print(
    tracer_ref.grid_at_redshift_from_image_plane_grid_and_redshift(
        image_plane_grid=image_plane_grid_stack.regular, redshift=1.0
    )
)

tracer_test = ray_tracing.Tracer(
    galaxies=[lens_galaxy, mass_sheet, source_galaxy],
    image_plane_grid_stack=image_plane_grid_stack,
    cosmology=cosmo,
)

print(
    tracer_test.grid_at_redshift_from_image_plane_grid_and_redshift(
        image_plane_grid=image_plane_grid_stack.regular, redshift=1.0
    )
)



import os

import autofit as af
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import mask as msk
from autolens.pipeline import pipeline as pl

# Welcome to the pipeline runner. This tool allows you to load strong lens instrument, and pass it to pipelines for a
# PyAutoLens analysis. To show you around, we'll load up some example instrument and run it through some of the example
# pipelines that come distributed with PyAutoLens.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out different lens
# names and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# It is convenient to specify the instrument type and data name as a string, so that if the pipeline is applied to multiple
# images we don't have to change all of the path entries in the load_ccd_data_from_fits function below.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = workspace_path + "../../outputs/PyAutoLens/"
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=output_path
)

data_type = "rjlens"
data_name = "f814w"

pixel_scale = 0.04

# Create the path where the instrument will be loaded from, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
    renormalize_psf=True,
    resized_ccd_shape=(251, 251),
    resized_psf_shape=(15, 15),
)

mask = msk.Mask.circular(
    shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.7
)

positions = abstract_data.load_positions(positions_path=data_path + "positions.dat")

ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data, mask=mask, positions=positions)

stop

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the lens instrument to its run function.
# Below, we'll' use a 3 phase example pipeline to fit the instrument with a parametric lens light, mass and source light
# profile. Checkout 'workspace/pipelines/examples/lens_light_and_x1_source_parametric.py' for a full description of
# the pipeline.

# The phase folders input determines the output directory structure of the pipeline, for example the input below makes
# the directory structure:
# 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/' or
# 'autolens_workspace/output/example/lens_light_and_x1_source/lens_light_and_x1_source_parametric/'

# For large samples of images, we can therefore easily group lenses that are from the same sample or modeled using the
# same pipeline.

pipeline_settings = pl.PipelineSettings(include_shear=True, fix_lens_light=False)

from workspace_jam.pipelines.advanced.with_lens_light.sersic.initialize import (
    lens_sersic_sie__source_sersic,
)


pipeline_initialize = lens_sersic_sie__source_sersic.make_pipeline(
    pipeline_settings=pipeline_settings,
    phase_folders=[data_type, data_name],
    positions_threshold=0.5,
)

pipeline = pipeline_initialize
pipeline.run(data=ccd_data, mask=mask, positions=positions)



from autofit.tools import path_util
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.lens.plotters import ray_tracing_plotters
from autolens.data.plotters import ccd_plotters

import os

# This tool allows one to make simulated data-sets of strong lenses, which can be used to test example pipelines and
# investigate strong lens modeling on data-sets where the 'true' answer is known.

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))

# The 'data name' is the name of the data folder and 'data_name' the folder the instrument is stored in, e.g:

# The image will be output as '/workspace/data/data_type/data_name/image.fits'.
# The noise-map will be output as '/workspace/data/data_type/data_name/lens_name/noise_map.fits'.
# The psf will be output as '/workspace/data/data_type/data_name/psf.fits'.

# (these files are already in the workspace and are remade running this script)
data_type = "example"
data_name = "lens_light_and_x1_source"

# Create the path where the instrument will be output, which in this case is
# '/workspace/data/example/lens_light_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

# The pixel scale of the image to be simulated
pixel_scale = 0.1

# Simulate a simple Gaussian PSF for the image.
psf = abstract_data.PSF.from_gaussian(
    shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale
)

# Setup the image-plane grid stack of the CCD array which will be used for generating the image-plane image of the
# simulated strong lens. The sub-grid size of 20x20 ensures we fully resolve the central regions of the lens and source
# galaxy light.
image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=pixel_scale, sub_grid_size=16
)

# Setup the lens galaxy's light (elliptical Sersic), mass (SIE+Shear) and source galaxy light (elliptical Sersic) for
# this simulated lens.
lens_galaxy = g.Galaxy(
    redshift=0.5,
    light=lp.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.9,
        phi=45.0,
        intensity=1.0,
        effective_radius=0.8,
        sersic_index=4.0,
    ),
)


# Use these galaxies to setup a tracer, which will generate the image-plane image for the simulated CCD instrument.
tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
    galaxies=[lens_galaxy], image_plane_grid_stack=image_plane_grid_stack
)

# Simulate the CCD instrument, remembering that we use a special image-plane image which ensures edge-effects don't
# degrade our modeling of the telescope optics (e.g. the PSF convolution).
ccd = ccd.SimulatedCCDData.from_image_and_exposure_arrays(
    array=tracer.padded_profile_image_plane_image_2d_from_psf_shape,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=False,
)

# Lets plot the simulated CCD instrument before we output it to files.
ccd_plotters.plot_ccd_subplot(ccd_data=ccd)

# Finally, lets output our simulated instrument to the instrument path as .fits files.
ccd.output_ccd_data_to_fits(
    ccd_data=ccd,
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    overwrite=True,
)



from test.integration.tests.lens_only import lens_x1_galaxy
from test.integration.tests.lens_only import lens_x1_galaxy_hyper
from test.integration.tests.lens_only import lens_x1_galaxy_link_param
from test.integration.tests.runner import run_a_mock


class TestCase:
    def _test_lens_x1_galaxy(self):
        run_a_mock(lens_x1_galaxy)

    def _test_lens_x1_galaxy_hyper(self):
        run_a_mock(lens_x1_galaxy_hyper)

    def _test_lens_x1_galaxy_link_param(self):
        run_a_mock(lens_x1_galaxy_link_param)



import autofit as af
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import pipeline_tagging
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp


# In this pipeline, we'll demonstrate passing redshifts to a pipeline - which means that the results and images of this
# pipeline will be returned in physical units (e.g. lengths in kpcs as well as arcsec, luminosities in magnitudes,
# masses in solMass, etc).

# The redshift of the lens and source are input parameters of all pipelines, and they take default values of 0.5 and
# 1.0. Thus, *all* pipelines will return physical values assuming these fiducial values if no other values are
# specified. Care must be taken interpreting the distances and masses if these redshifts are not correct or if the
# true redshifts of the lens and / or source galaxies are unknown.

# We'll perform a basic analysis which fits a lensed source galaxy using a parametric light profile where
# the lens's light is omitted. This pipeline uses two phases:

# Phase 1:

# Description: Fit the lens mass model and source light profile using x1 source.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Inputs the pipeline default redshifts where the lens has redshift 0.5, source 1.0.

# Phase 1:

# Description: Fits the lens and source model again..
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: Manually over-rides the lens redshift to 1.0 and source redshift to 2.0, to illustrate the different results.


def make_pipeline(phase_folders=None, redshift_lens=0.5, redshift_source=1.0):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below. However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_feature__redshifts_lens_and_source"

    pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings()

    # Unlike other features, the redshifts of the lens and source do not change the settings tag and phase path. Thus,
    # our output will simply go to the phase path:

    # phase_path = 'phase_name/settings'

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/pipeline_name/pipeline_tag/phase_name/phase_tag//'

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return msk.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Use the input value of redshifts from the pipeline.

    class LensSourceX1Phase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens.mass.centre_0 = af.GaussianPrior(mean=0.0, sigma=0.1)
            self.galaxies.lens.mass.centre_1 = af.GaussianPrior(mean=0.0, sigma=0.1)

    phase1 = LensSourceX1Phase(
        phase_name="phase_1__x1_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalIsothermal,
                shear=mp.ExternalShear,
            ),
            source_0=gm.GalaxyModel(
                redshift=redshift_source, light=lp.EllipticalSersic
            ),
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Use manually specified new values of redshifts for the lens and source galaxies.

    class LensSourceX2Phase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens = results.from_phase(
                "phase_1__x1_source"
            ).variable.galaxies.lens

            self.galaxies.source = results.from_phase(
                "phase_1__x1_source"
            ).variable.galaxies.source

    phase2 = LensSourceX2Phase(
        phase_name="phase_2__x2_source",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=1.0, mass=mp.EllipticalIsothermal, shear=mp.ExternalShear
            ),
            source=gm.GalaxyModel(redshift=2.0, light=lp.EllipticalSersic),
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import pipeline_tagging
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# In this pipeline, we'll perform an analysis which fits an image with no lens light, and a source galaxy using an
# inversion, using a power-law mass profile. The pipeline follows on from the inversion pipeline
# ''pipelines/no_lens_light/inversion/lens_sie__source_inversion_from_initialize.py'.

# The pipeline is two phases, as follows:

# Phase 1:

# Description: Fits the lens mass model as a power-law, using an inversion for the Source.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiBrightnessImage + Constant
# Previous Pipelines: no_lens_light/inversion/lens_sie__source_inversion_from_initialize.py
# Prior Passing: Lens Mass (variable -> previous pipeline), source inversion (variable / constant -> previous pipeline)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations.


def make_pipeline(
    pipeline_settings,
    phase_folders=None,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    signal_to_noise_limit=None,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=None,
    use_inversion_border=True,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_power_law_hyper__lens_power_law__source_inversion"

    pipeline_tag = pipeline_tagging.pipeline_tag_from_pipeline_settings(
        hyper_galaxies=pipeline_settings.hyper_galaxies,
        hyper_image_sky=pipeline_settings.hyper_image_sky,
        hyper_background_noise=pipeline_settings.hyper_background_noise,
        include_shear=pipeline_settings.include_shear,
        pixelization=pipeline_settings.pixelization,
        regularization=pipeline_settings.regularization,
    )

    phase_folders.append(pipeline_name)
    phase_folders.append(pipeline_tag)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy mass using the EllipticalIsothermal fit of the previous pipeline, and
    #    source inversion of the previous pipeline.

    class LensSourcePhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            ### Lens Mass, SIE -> PL ###

            self.galaxies.lens.mass.centre = results.from_phase(
                "phase_4__lens_sie__source_inversion"
            ).variable.galaxies.lens.mass.centre

            self.galaxies.lens.mass.axis_ratio = results.from_phase(
                "phase_4__lens_sie__source_inversion"
            ).variable.galaxies.lens.mass.axis_ratio

            self.galaxies.lens.mass.phi = results.from_phase(
                "phase_4__lens_sie__source_inversion"
            ).variable.galaxies.lens.mass.phi

            self.galaxies.lens.mass.einstein_radius = (
                results.from_phase("phase_4__lens_sie__source_inversion")
                .variable_absolute(a=0.3)
                .galaxies.lens.mass.einstein_radius
            )

            ### Lens Shear, Shear -> Shear ###

            if pipeline_settings.include_shear:

                self.galaxies.lens.shear = results.from_phase(
                    "phase_4__lens_sie__source_inversion"
                ).variable.galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.galaxies.source.pixelization = results.from_phase(
                "phase_4__lens_sie__source_inversion"
            ).hyper_combined.constant.galaxies.source.pixelization

            self.galaxies.source.regularization = results.from_phase(
                "phase_4__lens_sie__source_inversion"
            ).hyper_combined.constant.galaxies.source.regularization

            ## Set all hyper_galaxy-galaxies if feature is turned on ##

            if pipeline_settings.hyper_galaxies:

                self.galaxies.source.hyper_galaxy = (
                    results.last.hyper_combined.constant.galaxies.source.hyper_galaxy
                )

            if pipeline_settings.hyper_image_sky:

                self.hyper_image_sky = (
                    results.last.hyper_combined.constant.hyper_image_sky
                )

            if pipeline_settings.hyper_background_noise:

                self.hyper_background_noise = (
                    results.last.hyper_combined.constant.hyper_background_noise
                )

    phase1 = LensSourcePhase(
        phase_name="phase_1__lens_power_law__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            ),
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pipeline_settings.pixelization,
                regularization=pipeline_settings.regularization,
            ),
        ),
        sub_grid_size=sub_grid_size,
        signal_to_noise_limit=signal_to_noise_limit,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        use_inversion_border=use_inversion_border,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=pipeline_settings.hyper_galaxies,
        include_background_sky=pipeline_settings.hyper_image_sky,
        include_background_noise=pipeline_settings.hyper_background_noise,
        inversion=True,
    )

    return pipeline.PipelineImaging(pipeline_name, phase1, hyper_mode=True)



import sys
import os
import re
import getdist
import numpy as np

sys.path.append("../autolens/")

import galaxy
from profiles import light_profiles, mass_profiles
from astropy import cosmology


def getModelFolders(dir):
    regex = re.compile(".*Hyper")
    folders = [
        name
        for name in os.listdir(dir)
        if os.path.isdir(dir + name) and re.match(regex, name) == None
    ]
    return folders


ltm_skip = 0
center_skip = 0

image_dir = (
    "/gpfs/data_vector/pdtw24/PL_Data/SL03_2/"
)  # Dir of Object to make evidence tables from

image_name = "SLACSJ0252+0039"
redshift = 0.2803
source_redshift = 0.9818
scale_r = 6.83924
einstein_radius = 1.02
source_light_min = 0.8
source_light_max = 1.3
slacs_mass = "{0:e}".format(10 ** 11.25)
slacs_radius = 4.4

image_name = "SLACSJ1250+0523"
redshift = 0.2318
source_redshift = 0.7953
scale_r = 7.8660
kpc_to_arc = 3.8138
einstein_radius = 1.120
source_light_min = 0.5
source_light_max = 1.5
slacs_mass = "{0:e}".format(10 ** 11.26)
slacs_radius = 4.18

# image_name = 'SLACSJ1430+4105'
# redshift = 0.2850
# source_redshift = 0.5753
# scale_r = 6.771
# kpc_to_arc = 4.4309
# einstein_radius = 1.468
# source_light_min = 0.2
# source_light_max = 2.3
# slacs_mass =  '{0:e}'.format(10**11.73)
# slacs_radius = 6.53

# pipeline_folder = 'Pipeline_Total'
# phase_folder = 'PL_Phase_8'

pipeline_folder = "Pipeline_Total"
phase_folder = "PL_Phase_8"

image_dir = image_dir + image_name + "/" + pipeline_folder + "/" + phase_folder + "/"

model_folder = getModelFolders(image_dir)

pdf_file = image_dir + model_folder[0] + "/" + image_name + ".txt"

pdf = getdist.mcsamples.loadMCSamples(pdf_file)

values = pdf.getMeans()
params = pdf.getParams()

upper_conf = []
lower_conf = []
for i in range(len(values)):
    upper_conf.append(pdf.confidence(paramVec=i, limfrac=0.999, upper=True))
    lower_conf.append(pdf.confidence(paramVec=i, limfrac=0.999, upper=False))

if image_name == "SLACSJ1250+0523" or image_name == "SLACSJ1430+4105":
    center_skip = 2

sersic_bulge = light_profiles.SersicLightProfile(
    centre=(values[0], values[1]),
    axis_ratio=values[5],
    phi=values[6],
    intensity=values[2],
    effective_radius=values[3],
    sersic_index=values[4],
)

setattr(sersic_bulge, "centre_lower_limit_3_sigma", (lower_conf[0], lower_conf[1]))
setattr(sersic_bulge, "axis_ratio_lower_limit_3_sigma", lower_conf[5])
setattr(sersic_bulge, "phi_lower_limit_3_sigma", lower_conf[6])
setattr(sersic_bulge, "intensity_lower_limit_3_sigma", lower_conf[2])
setattr(sersic_bulge, "effective_radius_lower_limit_3_sigma", lower_conf[3])
setattr(sersic_bulge, "sersic_index_lower_limit_3_sigma", lower_conf[4])

setattr(sersic_bulge, "centre_upper_limit_3_sigma", (upper_conf[0], upper_conf[1]))
setattr(sersic_bulge, "axis_ratio_upper_limit_3_sigma", upper_conf[5])
setattr(sersic_bulge, "phi_upper_limit_3_sigma", upper_conf[6])
setattr(sersic_bulge, "intensity_upper_limit_3_sigma", upper_conf[2])
setattr(sersic_bulge, "effective_radius_upper_limit_3_sigma", upper_conf[3])
setattr(sersic_bulge, "sersic_index_upper_limit_3_sigma", upper_conf[4])

exponential_halo = light_profiles.ExponentialLightProfile(
    axis_ratio=values[9 + center_skip],
    phi=values[10 + center_skip],
    intensity=values[7 + center_skip],
    effective_radius=values[8 + center_skip],
)

setattr(
    exponential_halo,
    "centre_lower_limit_3_sigma",
    (lower_conf[9 + center_skip], lower_conf[10 + center_skip]),
)
setattr(exponential_halo, "axis_ratio_lower_limit_3_sigma", lower_conf[9 + center_skip])
setattr(exponential_halo, "phi_lower_limit_3_sigma", lower_conf[10 + center_skip])
setattr(exponential_halo, "intensity_lower_limit_3_sigma", lower_conf[7 + center_skip])
setattr(
    exponential_halo,
    "effective_radius_lower_limit_3_sigma",
    lower_conf[8 + center_skip],
)
setattr(
    exponential_halo,
    "centre_upper_limit_3_sigma",
    (upper_conf[9 + center_skip], upper_conf[10 + center_skip]),
)
setattr(exponential_halo, "axis_ratio_upper_limit_3_sigma", upper_conf[9 + center_skip])
setattr(exponential_halo, "phi_upper_limit_3_sigma", upper_conf[10 + center_skip])
setattr(exponential_halo, "intensity_upper_limit_3_sigma", upper_conf[7 + center_skip])
setattr(
    exponential_halo,
    "effective_radius_upper_limit_3_sigma",
    upper_conf[8 + center_skip],
)

power_law_mass = mass_profiles.EllipticalPowerLawMassProfile(
    centre=(values[11 + center_skip], values[12 + center_skip]),
    einstein_radius=values[13 + center_skip],
    axis_ratio=values[14 + center_skip],
    phi=values[15 + center_skip],
    slope=values[16 + center_skip],
)

# power_law_mass = mass_profiles.EllipticalPowerLawMassProfile(origin=(values[10], values[11]),
#                                                              einstein_radius=values[12],
#                                                              axis_ratio=values[13], phi=values[14],
#                                                              slope=values[15])

setattr(
    power_law_mass,
    "centre_lower_limit_3_sigma",
    (upper_conf[11 + center_skip], upper_conf[12 + center_skip]),
)
setattr(
    power_law_mass, "einstein_radius_lower_limit_3_sigma", lower_conf[13 + center_skip]
)
setattr(power_law_mass, "axis_ratio_lower_limit_3_sigma", lower_conf[14 + center_skip])
setattr(power_law_mass, "phi_lower_limit_3_sigma", lower_conf[15 + center_skip])
setattr(power_law_mass, "slope_lower_limit_3_sigma", lower_conf[16 + center_skip])

setattr(
    power_law_mass,
    "centre_upper_limit_3_sigma",
    (upper_conf[11 + center_skip], upper_conf[12 + center_skip]),
)
setattr(
    power_law_mass, "einstein_radius_upper_limit_3_sigma", upper_conf[13 + center_skip]
)
setattr(power_law_mass, "axis_ratio_upper_limit_3_sigma", upper_conf[14 + center_skip])
setattr(power_law_mass, "phi_upper_limit_3_sigma", upper_conf[15 + center_skip])
setattr(power_law_mass, "slope_upper_limit_3_sigma", upper_conf[16 + center_skip])

lens_galaxy = galaxy.Galaxy(
    redshift=redshift,
    light_profiles=[sersic_bulge, exponential_halo],
    mass_profiles=[power_law_mass],
)

source_galaxy = galaxy.Galaxy(redshift=source_redshift)

galaxy.LensingPlanes(
    galaxies=[lens_galaxy, source_galaxy],
    cosmological_model=cosmology.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7),
)

radius = lens_galaxy.mass_profiles[0].einstein_radius_in_units

total_mass = lens_galaxy.mass_within_circles(radius=radius)

lens_galaxy.mass_profiles[0].centre = lens_galaxy.mass_profiles[
    0
].centre_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].einstein_radius_in_units = lens_galaxy.mass_profiles[
    0
].einstein_radius_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_lower_limit_3_sigma
lens_galaxy.mass_profiles[0].slope = lens_galaxy.mass_profiles[
    0
].slope_lower_limit_3_sigma

total_mass_lower = lens_galaxy.mass_within_circles(radius=radius)

lens_galaxy.mass_profiles[0].centre = lens_galaxy.mass_profiles[
    0
].centre_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].einstein_radius_in_units = lens_galaxy.mass_profiles[
    0
].einstein_radius_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].phi = lens_galaxy.mass_profiles[0].phi_upper_limit_3_sigma
lens_galaxy.mass_profiles[0].slope = lens_galaxy.mass_profiles[
    0
].slope_upper_limit_3_sigma

total_mass_upper = lens_galaxy.mass_within_circles(radius=radius)

# print(lens_galaxy.mass_profiles[0].origin)
# print(lens_galaxy.mass_profiles[0].einstein_radius*kpc_to_arc)
# print(lens_galaxy.mass_profiles[0].axis_ratio)
# print(lens_galaxy.mass_profiles[0].phi)
# print(lens_galaxy.mass_profiles[0].slope)
# stop

# print(lens_galaxy.critical_density)

# print(slacs_mass)
# print(slacs_radius)


# einstein_radius = slacs_radius*lens_galaxy.arcsec_per_kpc

print(einstein_radius)
print(
    "Our Mass = ",
    "{0:e}".format(total_mass),
    "(",
    "{0:e}".format(total_mass_upper),
    "{0:e}".format(total_mass_lower),
    ")",
)
print(float(slacs_mass) / total_mass)

stop

lens_galaxy.plot_density_as_function_of_radius(
    maximum_radius=5.0,
    number_bins=300,
    image_name=image_name,
    plot_errors=True,
    labels=[
        "Sersic Bulge",
        "EllipticalExponentialMass Stellar Halo",
        "Dark Matter Halo",
    ],
)



import autofit as af
from autolens.model.inversion import pixelizations as pix, regularization as reg
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from test.integration.tests import runner

test_type = "lens_and_source_inversion"
test_name = "lens_both_x1_source_x1_rectangular"
data_type = "lens_light__source_smooth"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class SourcePix(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens.mass.centre.centre_0 = 0.0
            self.galaxies.lens.mass.centre.centre_1 = 0.0
            self.galaxies.lens.mass.einstein_radius_in_units = 1.6
            self.galaxies.source.pixelization.shape_0 = 20.0
            self.galaxies.source.pixelization.shape_1 = 20.0

    phase1 = SourcePix(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5,
                light=lp.SphericalDevVaucouleurs,
                mass=mp.EllipticalIsothermal,
            ),
            source=gm.GalaxyModel(
                redshift=1.0, pixelization=pix.Rectangular, regularization=reg.Constant
            ),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(name, phase1)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids, mask
from autolens.model.inversion import inversions
from test.unit.mock.model import mock_inversion


class TestRegularizationTerm:
    def test__solution_all_1s__regularization_matrix_simple(self):

        matrix_shape = (3, 3)

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape=matrix_shape),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        inv.pixelization_values = np.array([1.0, 1.0, 1.0])

        inv.regularization_matrix = np.array(
            [[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        )

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [1.0, 1.0, 1.0] * [1.0, 1.0, 1.0] = [(1.0*1.0) + (1.0*0.0) + (1.0*0.0)] = [1.0, 1.0, 1.0]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*1.0) + (1.0*0.0)]
        #                             [1.0, 1.0, 1.0]   [(1.0*0.0) + (1.0*0.0) + (1.0*1.0)]

        # (s_T * H) * s = [1.0, 1.0, 1.0] * [1.0] = 3.0
        #                                   [1.0]
        #                                   [1.0]

        assert inv.regularization_term == 3.0

    def test__solution_and_regularization_matrix_range_of_values(self):

        matrix_shape = (3, 3)

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        # G_l term, Warren & Dye 2003 / Nightingale /2015 2018

        # G_l = s_T * H * s

        # Matrix multiplication:

        # s_T * H = [2.0, 3.0, 5.0] * [2.0,  -1.0,  0.0] = [(2.0* 2.0) + (3.0*-1.0) + (5.0 *0.0)] = [1.0, -1.0, 7.0]
        #                             [-1.0,  2.0, -1.0]   [(2.0*-1.0) + (3.0* 2.0) + (5.0*-1.0)]
        #                             [ 0.0, -1.0,  2.0]   [(2.0* 0.0) + (3.0*-1.0) + (5.0 *2.0)]

        # (s_T * H) * s = [1.0, -1.0, 7.0] * [2.0] = 34.0
        #                                    [3.0]
        #                                    [5.0]

        inv.pixelization_values = np.array([2.0, 3.0, 5.0])

        inv.regularization_matrix = np.array(
            [[2.0, -1.0, 0.0], [-1.0, 2.0, -1.0], [0.0, -1.0, 2.0]]
        )

        assert inv.regularization_term == 34.0


class TestLogDetMatrix:
    def test__determinant_of_positive_definite_matrix_via_cholesky(self):

        matrix_shape = (3, 3)

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        matrix = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        assert log_determinant == pytest.approx(
            inv.log_determinant_of_matrix_cholesky(matrix), 1e-4
        )

    def test__determinant_of_positive_definite_matrix_2_via_cholesky(self):

        matrix_shape = (3, 3)

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        matrix = np.array([[2.0, -1.0, 0.0], [-1.0, 2.0, -1.0], [0.0, -1.0, 2.0]])

        log_determinant = np.log(np.linalg.det(matrix))

        assert log_determinant == pytest.approx(
            inv.log_determinant_of_matrix_cholesky(matrix), 1e-4
        )

    def test__matrix_not_positive_definite__raises_reconstruction_exception(self):

        matrix_shape = (3, 3)

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        matrix = np.array([[2.0, 0.0, 0.0], [-1.0, 2.0, -1.0], [0.0, -1.0, 0.0]])

        with pytest.raises(exc.InversionException):
            assert pytest.approx(inv.log_determinant_of_matrix_cholesky(matrix), 1e-4)


class TestReconstructedDataVectorAndImage:
    def test__solution_all_1s__simple_blurred_mapping_matrix__correct_reconstructed_image(
        self
    ):

        matrix_shape = (3, 3)

        msk = mask.Mask(
            array=np.array(
                [[True, True, True], [False, False, False], [True, True, True]]
            ),
            pixel_scale=1.0,
        )

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=msk, sub_grid_size=1, psf_shape=(1, 1)
        )

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape, grid_stack),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        inv.pixelization_values = np.array([1.0, 1.0, 1.0, 1.0])

        inv.blurred_mapping_matrix = np.array(
            [[1.0, 1.0, 1.0, 1.0], [1.0, 0.0, 1.0, 1.0], [1.0, 0.0, 0.0, 0.0]]
        )
        # CCD pixel 0 maps to 4 pixs pixxels -> value is 4.0
        # CCD pixel 1 maps to 3 pixs pixxels -> value is 3.0
        # CCD pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (inv.reconstructed_data_1d == np.array([4.0, 3.0, 1.0])).all()
        assert inv.reconstructed_data_2d == np.array(
            [[0.0, 0.0, 0.0], [4.0, 3.0, 1.0], [0.0, 0.0, 0.0]]
        )

        assert inv.pixelization_errors_with_covariance == pytest.approx(
            np.array([[0.7, -0.3, -0.3], [-0.3, 0.7, -0.3], [-0.3, -0.3, 0.7]]), 1.0e-4
        )
        assert inv.pixelization_errors == pytest.approx(
            np.array([0.7, 0.7, 0.7]), 1.0e-4
        )

    def test__solution_different_values__simple_blurred_mapping_matrix__correct_reconstructed_image(
        self
    ):

        matrix_shape = (3, 3)

        msk = mask.Mask(
            array=np.array(
                [[True, True, True], [False, False, False], [True, True, True]]
            ),
            pixel_scale=1.0,
        )

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=msk, sub_grid_size=1, psf_shape=(1, 1)
        )

        inv = inversions.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=np.ones(9),
            noise_map_1d=np.ones(9),
            convolver=mock_inversion.MockConvolver(matrix_shape),
            mapper=mock_inversion.MockMapper(matrix_shape, grid_stack),
            regularization=mock_inversion.MockRegularization(matrix_shape),
        )

        inv.pixelization_values = np.array([1.0, 2.0, 3.0, 4.0])

        inv.blurred_mapping_matrix = np.array(
            [[1.0, 1.0, 1.0, 1.0], [1.0, 0.0, 1.0, 1.0], [1.0, 0.0, 0.0, 0.0]]
        )

        # # CCD pixel 0 maps to 4 pixs pixxels -> value is 1.0 + 2.0 + 3.0 + 4.0 = 10.0
        # # CCD pixel 1 maps to 3 pixs pixxels -> value is 1.0 + 3.0 + 4.0
        # # CCD pixel 2 maps to 1 pixs pixxels -> value is 1.0

        assert (inv.reconstructed_data_1d == np.array([10.0, 8.0, 1.0])).all()
        assert inv.reconstructed_data_2d == np.array(
            [[0.0, 0.0, 0.0], [10.0, 8.0, 1.0], [0.0, 0.0, 0.0]]
        )

        assert inv.pixelization_errors_with_covariance == pytest.approx(
            np.array([[0.7, -0.3, -0.3], [-0.3, 0.7, -0.3], [-0.3, -0.3, 0.7]]), 1.0e-4
        )
        assert inv.pixelization_errors == pytest.approx(
            np.array([0.7, 0.7, 0.7]), 1.0e-4
        )


#
# class TestPixelizationQuantities:
#
#     def test__compare_to_inversion_utils(self):
#
#         matrix_shape = (3,3)
#
#         msk = mask.Mask(array=np.array([[True, True, True],
#                                         [False, False, False],
#                                         [True, True, True]]), pixel_scale=1.0)
#
#         grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
#             mask=msk, sub_grid_size=1, psf_shape=(1,1))
#
#         inv = inversions.Inversion(
#             image_1d=np.ones(9), noise_map_1d=np.ones(9), convolver=mock_inversion.MockConvolver(matrix_shape),
#             mapper=mock_inversion.MockMapper(matrix_shape, grid_stack),
#             regularization=mock_inversion.MockRegularization(matrix_shape))
#
#         inv.pixelization_values = np.array([1.0, 1.0, 1.0, 1.0])
#
#         inv.blurred_mapping_matrix = np.array([[1.0, 1.0, 1.0, 1.0],
#                                                [1.0, 0.0, 1.0, 1.0],
#                                                [1.0, 0.0, 0.0, 0.0]])
#
#         pixelization_residuals_util = \
#             inversion_util.pixelization_residuals_from_pixelization_values_reconstructed_data_1d_and_mapping_quantities(
#                 pixelization_values=inv.pixelization_values, reconstructed_data_1d=inv.reconstructed_data_1d,
#                 sub_to_regular=inv.mapper.sub_to_regular, pixelization_to_sub_all=inv.mapper.pixelization_to_sub_all)



from autolens.data.array import grids
from autolens.data.instrument import interferometer
from autolens.data import fourier_transform as ft

import numpy as np
import time
import os

test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))
data_path = test_path + "data/uvplane/"

interferometer_data = interferometer.load_interferometer_data_from_fits(
    image_path=data_path + "image.fits",
    pixel_scale=0.05,
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    primary_beam_path=data_path + "primary_beam.fits",
    real_visibilities_path=data_path + "real_visibilities.fits",
    imaginary_visibilities_path=data_path + "imaginary_visibilities.fits",
    visibilities_noise_map_path=data_path + "visibilities_noise_map.fits",
    u_wavelengths_path=data_path + "u_wavelengths.fits",
    v_wavelengths_path=data_path + "v_wavelengths.fits",
    renormalize_psf=False,
    renormalize_primary_beam=False,
)

print(interferometer_data.uv_wavelengths)
stop

repeats = 1

visibilities = len(interferometer_data.visibilities)

grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
    shape=interferometer_data.image.shape,
    pixel_scale=interferometer_data.pixel_scale,
    sub_grid_size=1,
)
grid_radians = grid.in_radians

image_pixels = grid.shape[0] * grid.shape[1]
source_pixels = 1000

shape_data = 8 * visibilities
shape_preloads = visibilities * image_pixels * 2
shape_mapping_matrix = visibilities * source_pixels

total_shape = shape_data + shape_preloads + shape_mapping_matrix

print("Data Memory Use (GB) = " + str(shape_data * 8e-9))
print("PreLoad Memory Use (GB) = " + str(shape_preloads * 8e-9))
print("Mapping Matrix Memory Use (GB) = " + str(shape_mapping_matrix * 8e-9))
print("Total Memory Use (GB) = " + str(total_shape * 8e-9))
print()

intensities_1d = np.ones(shape=(image_pixels))

transformer = ft.Transformer(
    uv_wavelengths=interferometer_data.uv_wavelengths, grid_radians=grid_radians
)

print(transformer.preload_real_transforms[0, :])
print(transformer.preload_real_transforms[1, :])
print(transformer.preload_real_transforms[2, :])
stop


start = time.time()
for i in range(repeats):
    transformer.real_visibilities_from_intensities(intensities_1d=intensities_1d)
diff = time.time() - start
print("Real Visibilities Time = {}".format(diff / repeats))

start = time.time()
for i in range(repeats):
    transformer.real_visibilities_via_preload_from_intensities(
        intensities_1d=intensities_1d
    )
diff = time.time() - start
print("Real Visibilities PreLoad Time = {}".format(diff / repeats))
